#
#  APRS4R - a ruby based aprs gateway/digipeater
#  Copyright (C) 2006 by Michael Conrad <do5mc@friggleware.de>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
#

require 'thread'

require 'aprs4r/APRS4RLogger'

require 'aprs4r/GatewayPlugin'
require 'aprs4r/SocketManager'

require 'aprs4r/BuddyGatewayPluginConfiguration'


module APRS4R

  class BuddyGatewayPlugin < GatewayPlugin

    @logger = APRS4RLogger.get_logger( "BuddyGatewayPlugin")


    def initialize( configuration)
      logger.info( "initialize( configuration)")

      super( configuration)

      @buddies = []
      @buddies = configuration.buddies if configuration.buddies

      @call = ""
      @call = configuration.call if configuration.call

      @path = []
      @path = configuration.path if configuration.path

      begin
        @timeout = 300
        @timeout = configuration.timeout.to_i if configuration.timeout
      rescue Exception => ex
        logger.error( "Error reading timeout variable: #{configuration.timeout}")
      end

      @mutex = Mutex.new
      @stations = Hash.new
      @filter_calls = Array.new
      @filter_regexps = Array.new

      @filter_timeout = 12 * 60 * 60 # 12 hours
      @filter_stations = Hash.new

      # add buddies to filter
      @buddies.each{ |buddy|
        add_filter( buddy, nil)
      }

      update_filter

      return
    end


    def register_listener
      logger.info( "register_listener")

      if @enable
        SocketManager.add_listener( @inDevice, self)
        SocketManager.add_listener( @outDevice, self)
      end

      return
    end


    def unregister_listener
      logger.info( "unregister_listener")

      if @enable
        SocketManager.remove_listener( @inDevice, self)
        SocketManager.remove_listener( @outDevice, self)
      end

      return
    end


    def recv_message( name, message)
      logger.info( "recv_message( name, message)")
      
      if message.nil?
        return
      end

      if name == @inDevice
        result = handle_message( name, message)
      
        # station is in filter
        if result
          expire_stations

          # but station was heard locally
          if @stations.has_key?( message.source)
            logger.debug( "station was heard locally")
            return
          end

          # replace path and sending beacon in outDevice
          result.path = @path

          logger.log( "gating buddy from #{name} to #{@outDevice}, message: #{result}")
          SocketManager.send_message( @outDevice, result)
        end

      elsif name == @outDevice
        result = handle_message( name, message)

        # station is in filter
        if result 
          expire_stations

          # store station call of locally heard station
          @stations[message.source] = Time.now
        elsif message.is_message? && message.message_recipient == @call

          if message.message_ack
            ack = APRSMessage.new( @call, APRS4RCall)
            ack.payload = ":" + message.source[0..8].ljust( 9) + ":" + "ack" + message.message_ack

            SocketManager.send_message( @outDevice, ack)
          end
          
          payload = message.message_payload

          if payload
            
            if payload =~ /^\?APRSB/i
              # query buddy call
              values = payload.split( / /)

              if values.length >= 2
                buddy_call = values[1]

                response = APRSMessage.new( @call, APRS4RCall)
                response_text = nil

                if in_filter?( buddy_call)
                  # send positive response
                  logger.debug( "buddy: #{buddy_call} already in buddy list")

                  response_text = "BUDDY #{buddy_call} is already in buddy list."
                else
                  # send negative response
                  logger.debug( "buddy: #{buddy_call} not in buddy list")

                  response_text = "BUDDY #{buddy_call} is not in buddy list"
                end

                response.payload = ":" + message.source[0..8].ljust( 9) + ":" + response_text

                SocketManager.send_message( @outDevice, response)
              end

            elsif payload =~ /^\!APRSB/i
              # set buddy call
              values = payload.split( / /)

              if values.length >= 2
                buddy_call = values[1]

                remove_filter( buddy_call)
                add_filter( buddy_call)
                update_filter
                
                logger.debug( "buddy: #{buddy_call} added to buddy list")
                response_message = "BUDDY ${buddy_call} added to buddy list (12h)"

                response = APRSMessage.new( @call, APRS4RCall)
                response.payload = ":" + message.source[0..8].ljust( 9) + ":" + response_message

                SocketManager.send_message( @outDevice, response)

              end
            end
          end
        end
        
      else
        logger.warn( "listener call from wrong interface")
      end
      
      return
    end
  
  
    def handle_message( name, message)
      logger.info( "handle_message( name, message)")
      
      # NOGATE check
      path = message.path
      if path.nil? || path.include?( "NOGATE") || path.include?( "RFONLY")
        return
      end
      
      if in_filter?( message.source)
        return message
      end

      return
    end


    def expire_stations
      logger.info( "expire_stations")

      now = Time.now

      @mutex.synchronize do 
        # clean up local station list 
        @stations.each{ |key, timestamp|
          if now - timestamp > @timeout
            @stations.delete( key)
          end
        }

        # clean up filter station list 
        @filter_stations.each{ |key, timestamp|
          if timestamp && (now - timestamp > @filter_timeout)
            remove_filter( key)
          end
        }
      end
      
      return
    end


    def add_filter( buddy_call, time = Time.now)
      logger.info( "add_filter( buddy_call)")
      
      buddy_regexp = /#{buddy_call.gsub( /\*/, ".*")}/

      @filter_stations.store( buddy_call, time)
      @filter_calls.push( buddy_call)
      @filter_regexps.push( buddy_regexp)

      return
    end
    
    
    def remove_filter( buddy_call)
      logger.info( "remove_filter( buddy_call)")
      
      buddy_regexp = /#{buddy_call.gsub( /\*/, ".*")}/

      @filter_stations.delete( buddy_call)
      @filter_calls.delete( buddy_call)
      @filter_regexps.delete( buddy_regexp)

      return
    end
    
    
    def in_filter?( buddy)
      
      @filter_regexps.each{ |filter|
        if buddy =~ filter
          return true
        end
      }
      
      return false
    end


    def update_filter
      logger.info( "update_filter")

      logger.debug( "updating #{@inDevice} filter: #{@filter_calls}")
      SocketManager.update_filter( @inDevice, @filter_calls)

      return
    end

  end

end
