#
#  APRS4R - a ruby based aprs gateway/digipeater
#  Copyright (C) 2006 by Michael Conrad <do5mc@friggleware.de>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
#

require 'aprs4r/APRS4RLogger'

require 'aprs4r/APRSMessage'
require 'aprs4r/Plugin'
require 'aprs4r/QueryPluginConfiguration'


module APRS4R

  class QueryPlugin < Plugin

    @logger = APRS4RLogger.get_logger( "QueryPlugin")

    
    def initialize( configuration)
      logger.info( "initialize( configuration)")

      super( configuration)

      @device = configuration.device
      @call = /#{configuration.call}/

      # wildcard call
      if configuration.call =~ /\*$/
        @call = /#{configuration.call.gsub( /\*/, ".*")}/
      end

      @timeout = 900
      begin
        @timeout = configuration.timeout.to_i if configuration.timeout
      rescue Excpetion => ex
        logger.warn( "error converting timeout: #{ex}")
      end

      @position_message = nil
      @status_message = nil
      @object_message = nil
      @igate_message = nil
      @weather_message = nil

      @direct_stations = Hash.new
      @heard_stations = Hash.new
      @last_heard_stations = Array.new

      @message_stations = Hash.new

      @last_direct_expire = Time.now
      @last_heard_expire = Time.now

      return
    end
    
    
    def start 
      logger.info( "start")

      if ! @enable
        return
      end

      register_listener( @device)

      return
    end

    
    def register_listener( device)
      logger.info( "register_listener")

      if @enable
        SocketManager.add_listener( device, self)
        SocketManager.add_local_listener( device, self)
      end

      return
    end


    def unregister_listener( device)
      logger.info( "unregister_listener")

      SocketManager.remove_listener( device, self)
      SocketManager.remove_local_listener( device, self)

      return
    end


    def recv_message( name, message)
      logger.info( "recv_message( name, message)")

      if message.nil?
        return
      end

      results = handle_message( name, message)

      if results
        results.each{ |result|
          logger.log( "sending query result on #{@device}, result: #{result}")

          SocketManager.send_message( name, result)
        }
      end

      return
    end


    def handle_message( name, message)
      logger.info( "handle_message( name, message)")

      return if message.nil?

      return if name != @device

      results = nil

      if message.is_query?
        # handle general queries
        logger.debug( "general query")
        
        type = message.query_type

        # general queries
        case type

        when "APRS"
          results = handle_APRS( name, message)
        when "IGATE"
          results = handle_IGATE( name, message)
        when "WX"
          results = handle_WX( name, message)
        end
        
      elsif message.is_message? && message.message_recipient =~ @call
        # handle directed queries
        logger.debug( "directed query")

        type = message.query_type

        case type

        when "APRSD"
          results = handle_APRSD( name, message)

        when "APRSH"
          results = handle_APRSH( name, message) 
          
        when "APRSM"
          results = handle_APRSM( name, message)

        when "APRSO"
          results = handle_APRSO( name, message)

        when "APRSP"
          results = handle_APRSP( name, message)

        when "APRSS"
          results = handle_APRSS( name, message)
          
        when "APRST" 
          results = handle_APRST( name, message)

        when "PING?"
          results = handle_APRST( name, message)
          
        end

      else
        # handle all other aprs messages
        logger.debug( "other aprs message")

        if message.is_direct?
          @direct_stations[message.source] = Time.now

          if @last_direct_expire + @timeout < Time.now
            expire_direct_stations
            @last_direct_expire = Time.now
          end
        end

        # TODO flexible heard list 
        timestamps = @heard_stations[message.source]

        if timestamps.nil?
          timestamps = [] 
          @heard_stations[message.source] = timestamps
        end

        timestamps << Time.now

        @last_heard_stations.delete( message.source)
        @last_heard_stations.unshift( message.source)


        if message.is_message?
          logger.debug( "incoming message: #{message}")
          messages = @message_stations[message.message_recipient]

          if messages.nil?
            messages = []
            @message_stations[message.message_recipient] = messages
          end
          
          messages << message
        end


        if @last_heard_expire + @timeout < Time.now
          expire_heard_stations
          @last_heard_expire = Time.now
        end
      end
      
      return results
    end


    def recv_local_message( name, message)
      logger.info( "recv_local_message( name, message)")

      # fill messages
      if message.source.to_s =~ @call

        if message.has_position?
          @position_message = message.clone 
          logger.debug( "position: #{@position_message}")
        end

        if message.has_status?
          @status_message = message.clone
          logger.debug( "status: #{@status_message}")
        end

        if message.is_object?
          @object_message = message.clone
          logger.debug( "object: #{@object_message}")
        end

        if message.has_capabilities?( "IGATE")
          @igate_message = message.clone 
          logger.debug( "igate: #{@igate_message}")
        end
        
        if message.has_weather?
          @wx_message = message.clone 
          logger.debug( "wx: #{@wx_message}")
        end

      end

      return
    end


    def handle_APRS( name, message)
      logger.info( "handle_APRS( name, message)")

      results = Array.new

      # send position report
      results.concat( handle_APRSP( name, message))

      # send status report
      results.concat( handle_APRSS( name, message))

      return results
    end

    
    def handle_IGATE( name, message)
      logger.info( "handle_IGATE( name, message)")

      results = Array.new

      # send igate capabilities
      if @igate_message
        @igate_message.path = []
        results << @igate_message
      end

      return results
    end


    def handle_WX( name, message)
      logger.info( "handle_WX( name, message)")
      
      results = Array.new

      # send position report
      results.concat( handle_APRSP( name, message))

      # send wx report
      if @wx_message
        @wx_message.path = []
        results << @wx_message
      end

      return results
    end

    
    def handle_APRSD( name, message)
      logger.info( "handle_APRSD( name, message)")
      
      results = Array.new

      stations = String.new
      @direct_stations.each_key{ |station|
        stations += " #{station}"
      }

      response = APRSMessage.new( @call, APRS4RCall)
      response_text = "Directs=#{stations}"
      response.payload = ":" + message.source[0..8].ljust( 9) + ":" + response_text

      results << response

      return results
    end


    def handle_APRSH( name, message)
      logger.info( "handle_APRSH( name, message)")

      results = Array.new

      call = message.query_payload
      timestamps = @heard_stations[call]

      if !timestamps.nil? && !timestamps.empty?
        response = APRSMessage.new( @call, APRS4RCall)
        response_text = "#{call} HEARD:"

        now = Time.now
        hours = Array.new( 8, 0)

        timestamps.each{ |timestamp|
          index = (now - timestamp) % (60 * 60)

          hours[index] += 1 if index >= 0 && index < 8
        }

        hours.each{ |hour|
          if hour > 0 
            response_text += " #{hour}"
          else
            response_text += " ."
          end
        }

        response.payload = ":" + message.source[0..8].ljust( 9) + ":" + response_text

        results << response
      end

      return results
    end


    def handle_APRSM( name, message)
      logger.info( "handle_APRSM( name, message)")

      results = Array.new
      messages = @message_stations[message.source]

      if !messages.nil? && !messages.empty?
        messages.each{ |message|
          results << message.clone
        }
      end
        
      return results
    end


    def handle_APRSO( name, message)
      logger.info( "handle_APRSO( name, message)")

      results = Array.new

      if @object_message
        # TODO path handling
        @object_message.path = []
        results << @object_message
      end

      return results
    end


    def handle_APRSP( name, message)
      logger.info( "handle_APRSP( name, message)")

      results = Array.new
    
      if @position_message
        # TODO path handling
        @position_message.path = []
        results << @position_message
      end

      return results
    end
      

    def handle_APRSS( name, message)
      logger.info( "handle_ARPSS( name, message)")

      results = Array.new

      if @status_message
        # TODO path handling
        @status_message.path = []
        results << @status_message
      end

      return results
    end


    def handle_APRST( name, message)
      logger.info( "handle_APRST( name, message)")

      results = Array.new

      result = APRSMessage.new
      result.source = message.destination
      result.destination = message.source

      # TODO path handling
      result.path = ["WIDE3-3"]

      payload = sprintf( ":%9.9s:%s>%s", message.source.ljust( 9), message.destination, message.path.join( ','))

      result.payload = payload

      results << result
      
      return results
    end

  end

end
