#
#  APRS4R - a ruby based aprs gateway/digipeater
#  Copyright (C) 2006 by Michael Conrad <do5mc@friggleware.de>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
#

require 'aprs4r/APRS4RBase'
require 'aprs4r/APRS4RLogger'

require 'aprs4r/SocketRecvThread'
require 'aprs4r/SocketSendThread'


module APRS4R

  class SocketManager < APRS4RBase

    @logger = APRS4RLogger.get_logger( "SocketManager")


    @@singleton = nil

    @@recvThreads = Hash.new
    @@sendThreads = Hash.new

    @@sockets = Hash.new
    @@recv_listeners = Hash.new
    @@local_listeners = Hash.new
    @@duplicates = Hash.new

    @@recvCount = Hash.new
    @@sendCount = Hash.new
    @@lastMessage = nil


    def SocketManager.init( configuration)
      logger.info( "init( configuration)")

      sockets = configuration.devices

      if sockets.nil?
        return
      end

      begin

        sockets.each { |name, configuration|
          
          logger.debug( "configuration: #{configuration}")
          
          name = configuration.name
          type = configuration.type
          enable = configuration.enable
          
          # load socket
          if enable 
            logger.debug( "name: #{name}")
            logger.debug( "type: #{type}")
            logger.debug( "configuration: #{configuration}")
            
            socket = nil
            socket_class = "aprs4r" + "/" + type
            
            begin 
              require socket_class
              
              socket = eval( "#{type}.new( configuration)")
              
            rescue Exception => ex 
              logger.error( "Unable to load/initialize socket #{type}: #{ex}")
              socket = nil
            end


            if socket 
              logger.debug( "socket: #{socket}")
              
              # start recv thread
              recvThread = SocketRecvThread.new( socket)
              @@recvThreads[name] = recvThread
              
              Thread.new { 
                recvThread.run
              }
              
              logger.debug( "recv thread: #{recvThread}")
              
              # start send thread
              sendThread = SocketSendThread.new( socket)
              @@sendThreads[name] = sendThread
              Thread.new { 
                sendThread.run
              }
              
              logger.debug( "send thread: #{sendThread}")
              
              # start packet handle thread
              Thread.new { 
                run( name)
              }
              
              @@sockets[name] = socket
              @@recv_listeners[name] = Array.new
              @@local_listeners[name] = Array.new
              @@duplicates[name] = Hash.new
            
              @@recvCount[name] = 0
              @@sendCount[name] = 0
            end
          end
        }
        
      rescue Exception => ex
        logger.error( "Initialisation failed")
      end
      
      return
    end


    def SocketManager.run( name)
      logger.info( "run( name)")

      recvThread = @@recvThreads[name]

      while true 
        message = recvThread.recv_message

        SocketManager.recv_message( name, message)
      end

    end

    
    def SocketManager.add_listener( name, listener)
      logger.info( "add_listener( name, listener)")
      
      SocketManager.add_recv_listener( name, listener)
    end


    def SocketManager.add_recv_listener( name, listener)
      logger.info( "add_recv_listener( name, listener)")
      
      if @@sockets.key?( name) 
        listeners = @@recv_listeners[name]
        listeners << listener
      end

      return
    end


    def SocketManager.remove_listener( name, listener)
      logger.info( "remove_listener( name, listener)")

      SocketManager.remove_recv_listener( name, listener)
    end


    def SocketManager.remove_recv_listener( name, listener)
      logger.info( "remove_recv_listener( name, listener)")

      if @@sockets.key?( name)
        listeners = @@recv_listeners[name]
        listeners.delete( listener)
      end

      return
    end


    def SocketManager.add_local_listener( name, listener)
      logger.info( "add_local_listener( name, listener)")

      if @@sockets.key?( name)
        listeners = @@local_listeners[name]
        listeners << listener
      end

      return
    end


    def SocketManager.remove_local_listener( name, listener)
      logger.info( "remove_local_listener( name, listener)")

      if @@sockets.key?( name)
        listeners = @@local_listeners[name]
        listeners.delete( listener)
      end

      return
    end


    def SocketManager.recv_message( name, message)
      logger.info( "recv_message( name, message)")

      if message.nil?
        return
      end

      if SocketManager.duplicate?( name, message)
        logger.log( "recv: duplicate message dropped")
        return
      end

      @@recvCount[name] += 1
      @@lastMessage = message
      
      # deliver message to all registered listeners
      listeners = @@recv_listeners[name]
      
      if ! listeners.nil?
        listeners.each{ |listener|
          begin 
            listener.recv_message( name, message.clone)
          rescue Exception => ex 
            logger.error( "recv_message failed for plugin: #{listener}, ex: #{ex}")
          end
        }
      end

      return
    end

    
    def SocketManager.send_message( name, message)
      logger.info( "send_message( name, message)")
      
      sendThread = @@sendThreads[name]

      if message.nil? 
        logger.warn( "device: #{name}, message: NIL")
        return
      end
      
      if sendThread.nil?
        logger.warn( "device: #{name}, thread: NIL")
        return
      end

      if SocketManager.duplicate?( name, message)
        logger.log( "send: duplicate message dropped (from #{message.source})")
        return
      end

      @@sendCount[name] += 1

      # send message and wakeup waiting thread
      sendThread.send_message( name, message)

      SocketManager.add_duplicate( name, message)

      return
    end


    def SocketManager.send_local_message( name, message)
      logger.info( "send_local_message( name, message)")

      if message.nil?
        return
      end

      # deliver message to all registered listeners
      listeners = @@local_listeners[name]

      if ! listeners.nil?
        listeners.each{ |listener|
          begin 
            listener.recv_local_message( name, message.clone)
          rescue Exception => ex
            logger.error( "recv_local_message failed for plugin: #{listener}, ex: #{ex}")
          end
        }
      end

      return
    end


    def SocketManager.update_filter( name, filter)
      logger.info( "update_filter( filter)")

      socket = @@sockets[name]

      begin
        socket.update_filter( filter) if socket
      rescue Exception => ex
        logger.error( "filter update failed: #{ex}")
      end

      return
    end


    def SocketManager.add_duplicate( name, message, now = Time.now)
      logger.info( "add_duplicate( name, message, now)")

      duplicatePeriod = 0

      socket = @@sockets[name]

      if !socket.nil?
        duplicatePeriod = socket.duplicatePeriod
      end

      duplicates = @@duplicates[name]

      if duplicatePeriod
        # remove old entries from duplicate list
        duplicates.each{ |key, timestamp|
          if now - timestamp > duplicatePeriod
            duplicates.delete( key)
          end
        }
      end
      
      # add message to duplicate list
      duplicates[message.duplicate_key] = now

      return
    end


    def SocketManager.duplicate?( name, message, now = Time.now)
      logger.info( "duplicate?( name, message, now)")
      
      duplicatePeriod = 0

      socket = @@sockets[name]

      if !socket.nil?
        duplicatePeriod = socket.duplicatePeriod
      end

      # duplicate check (if enabled)
      if duplicatePeriod 
        duplicates = @@duplicates[name]
        
        timestamp = duplicates[message.duplicate_key]
        
        # drop if timestamp exists and message is a duplicate
        if timestamp && (now - timestamp < duplicatePeriod)
          return true
        end
      end

      return false
    end

    
    def SocketManager.recv_count( name)
      return @@recvCount[name]
    end

    def SocketManager.send_count( name)
      return @@sendCount[name]
    end

    def SocketManager.last_message
      return @@lastMessage
    end

  end

end
